/****************************************************/
/*                assignment3.c                     */
/*             Written by Tzu-Kai Chou              */
/****************************************************/

#include <stdio.h>
#include <math.h>
#include <mpi.h>

double sqrt();

int main(int argc, char *argv[])
{
    int num_intervals, myid, num_procs, i, j;
    double mypi_sq=0.0, pi_sq=0.0, h, x, y;
    
    FILE *fp;
    char str[81];

    MPI_Comm grid_comm;
    int dim_sizes[2]={1,1};
    int wrap_around[2]={1,1};
    int coordinates[2];
    int my_grid_rank; 

    double f(double x, double y);
    void My_SetGrid(int p, int dim[]);

/*******************************************************************/
/*  Initialization                                                 */
/*******************************************************************/
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);

/*******************************************************************/
/*  Program obtain user input data by reading the file input.dat   */
/*  Only process 0 go to read num_intervals then broadcast.        */
/*  Now we can make sure each process get the same num_intervals.  */
/*  For any unexpectible situation, default value is 1024.         */ 
/*******************************************************************/
    if(myid==0)
    {
         fp=fopen("input.dat","r");
         if(fp==NULL) /* file can't be opened or doesn't exist */
            num_intervals=1024;
         else
         {
            fgets(str, 80, fp);
            num_intervals=atoi(str);
            if(num_intervals<=0) /* illegal num_intervals value */
               num_intervals=1024; 
         }  
         fclose(fp);         
    }
    MPI_Bcast(&num_intervals,1,MPI_INT,0,MPI_COMM_WORLD);

/*******************************************************************/
/*  Determine dimention sizes.                                     */
/*  Create cartesian grid communicator.                            */
/*  Determine the coordinates of each process.                     */ 
/*******************************************************************/
    My_SetGrid(num_procs, dim_sizes);
    MPI_Cart_create(MPI_COMM_WORLD,2,dim_sizes,wrap_around,1,&grid_comm);
    MPI_Comm_rank(grid_comm, &my_grid_rank);
    MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);

/*******************************************************************/
/*  Each process compute the value of its own subinterval          */
/*  using 2D process rank.                                         */
/*******************************************************************/
    h=1.0/(double)num_intervals;
    for(i=coordinates[0];i<num_intervals;i+=dim_sizes[0])
        for(j=coordinates[1];j<num_intervals;j+=dim_sizes[1]) 
        { 
           x = i*h;
           y = j*h;
           mypi_sq += h*h*(f(x,y)+f(x,y+h)+f(x+h,y)+f(x+h,y+h));
        }
    mypi_sq = mypi_sq*4;

/*******************************************************************/
/*  Process 0 sum up the results to get the total vale             */
/*  using original 1D process rank.                                */
/*******************************************************************/
    MPI_Reduce(&mypi_sq,&pi_sq,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);

/*******************************************************************/
/*  Output the PI square on screen.                                */
/*******************************************************************/
    if(myid==0)
    {
      printf(" Number of intervals : %d\n",num_intervals);
      printf("     Number of nodes : %d\n",num_procs);
      printf("Cartesian grid sizes : %d X %d\n",dim_sizes[0],dim_sizes[1]);
      printf("           PI square : %.16f\n",pi_sq);  
      printf("           PI: %.16f\n", sqrt(pi_sq));  
    }

/*******************************************************************/
/*  Finalization                                                   */
/*******************************************************************/
    MPI_Finalize();

}


/*******************************************************************/
/*  Description : hard-wired function                              */
/*                f(x,y) = 1 /((1+x^2)*(1+y^2))                    */
/*        Input : x, y                                             */
/*        Output: f(x,y)                                           */
/*******************************************************************/
double f(double x, double y)
{
    double result;
    result=1.0/((1+x*x)*(1+y*y));   
    return result;
}

/*******************************************************************/
/*  Description : This function determine the dimention sizes.     */ 
/*        Input : Number of process,Arry of dimention sizes        */
/*       Output : N/A                                              */ 
/*                                                                 */
/*      example :    p       SIZE                                  */
/*                   1       1 X 1                                 */
/*                   2       1 X 2                                 */
/*                   3       1 X 3                                 */
/*                   4       2 X 2                                 */
/*                   5       1 X 5                                 */
/*                   6       3 X 2                                 */
/*                   7       1 X 7                                 */
/*                   8       4 X 2                                 */
/*******************************************************************/
void My_SetGrid(int p, int dim[])
{
    int i;
    for(i=1;i<=p/2;i++)
    {
        if((p%i)==0)
           dim[0]=i;   
    }
    dim[1]=p/dim[0];
}
